-- Windows interface for PragmAda Software Engineering's Mine Detector Game
-- using Claw (demo version, downloadable from www.rrsoftware.com)
-- This is a derived work under the Gnu General Public License, see below.
-- tmoran@acm.org 12/12/01
--
-- Mine Detector Game
-- Copyright (C) 2001 by PragmAda Software Engineering.  All rights reserved.
-- **************************************************************************
--
with Claw,
     Claw.Attributes,
     Claw.Brushes,
     Claw.Canvas,
     Claw.Basic_Window,
     Claw.Draw,
     Claw.Edit.Multiline,
     Claw.Fonts,
     Claw.Icons,
     Claw.Menus.Bar,
     Claw.Menus.Lists,
     Claw.Message_Box,
     Claw.Pens,
     Claw.Rectangles,
     Claw.Window_Operations;
with Field.Operations;

package body User_If is

  -- Game board assumed by callers of user_if.display, etc
  Board : Game_Board_Type;

  -- Cosmetic definitions

  Game_Over_Message : constant String := "GAME OVER";

  About_Text : constant String
    :=("  Mine Detector using CLAW" & Ascii.LF &
       "               based on" & Ascii.LF &
       "   Mine Detector, (C) 2001," & Ascii.LF &
       "PragmAda Software Engineering" & Ascii.LF &
       "            ""Ada Inside""");

  Blank_Line : constant String := ascii.cr & ascii.lf & ascii.cr & ascii.lf;

  Rules_Text : constant String
    :=   "The object of the game is to mark all cells containing "
       & "mines and to step on all cells that do not contain a "
       & "mine."
       & Blank_Line
       & "The game is played on a rectangular field of 16 x 30 "
       & "cells. 99 mines are hidden within the field. "
       & "Some of the cells have numbers on them. The numbers represent "
       & "the total number of mines in that cell and its "
       & "immediate neighbors. As you play the game, additional cells "
       & "will become numbered."
       & Blank_Line
       & "You step on a cell by clicking on it with the primary "
       & "mouse button (normally the left button when configured "
       & "for a right-handed user). You mark a cell by clicking "
       & "on it with the secondary mouse button (normally the "
       & "right button). A marked cell has an M on it.  Marking a "
       & "marked cell unmarks it. You can only mark or step "
       & "on a cell with a number on it - other cells do nothing."
       & Blank_Line
       & "When you step on a cell, an auto-stepping algorithm "
       & "automatically steps on any of its neighbors that "
       & "obviously do not contain mines. Since this is then "
       & "done for the neighbors of the stepped-on neighbors, "
       & "the auto-stepping algorithm will spread across areas "
       & "of the field that obviously do not contain mines. The "
       & "auto-stepping algorithm is invoked even if the cell is "
       & "already stepped on. This can be useful to clear around "
       & "a new mark."
       & Blank_Line
       & "If you step on a cell containing a mine, either "
       & "directly or indirectly through the auto-stepping "
       & "algorithm, flames erupt, and the game is over."
       & Blank_Line
       & "The game is over when you step on a mine, or when you "
       & "have marked all mines and stepped on all other cells. "
       & "When the game is over, 'GAME OVER' appears at the top."
       & Blank_Line
       & "At the top right of the window is a number. At the "
       & "start of a game this is 99, the number of mines in the "
       & "field. Each time you mark a cell, this number is "
       & "decreased by one. Each time you unmark a marked cell, "
       & "this number is increased by one. If you successfully "
       & "complete a game, this number will be zero."
       & Blank_Line
       & "The 'New Game' button starts a new game. Any game in "
       & "progress is abandoned."
       & Blank_Line
       & "The 'Auto Mark' option enables an auto-marking "
       & "algorithm that marks any cells that obviously contain "
       & "a mine. In this mode, the game does not present much "
       & "of an intellectual challenge."
       & Blank_Line
       & "The 'Auto Step after Mark' option enables the auto-"
       & "stepping algorithm after a cell is marked, either "
       & "directly or indirectly through the auto-marking "
       & "algorithm.";

  -- Pragma Resource is a Janus pragma to build .rc files.
  -- With a different compiler, just create a .rc file containing
  --   Fire ICON "fire.ico"
  Pragma Resource(ICON,"Fire", "fire.ico");
  --   Mark ICON "mark.ico"
  Pragma Resource(ICON,"Mark", "mark.ico");

  Fire_Icon : Claw.Icons.Icon_Type;    -- explosion
  Mark_Icon : Claw.Icons.Icon_Type;    -- mine warning flag

  Cell_Size : constant := 21;

  -- exported routines

  procedure Show_Game_Over is
  begin -- Show_Game_Over
    Claw.Rectangles.Invalidate(Board, Board.Game_Over_Area,
                               Erase_Background=>True);
    -- Note: we don't actually need to erase the background, since we will
    -- write all over it anyway, but it tends to cause a flashing effect
    -- that is nice to have when the game is done
  end Show_Game_Over;

  procedure Display (Cell : in Field.Cell_Location;
                     Text : in Cell_String;
                     Active : in Boolean) is
    use type Claw.Int;
    Top : Position_Lists renames Board.Cell_Top;
    Left : Position_Lists renames Board.Cell_Left;
  begin -- Display
    if Board.Button_State(Cell.Row, Cell.Column) /= (Text, Active) then
       Board.Button_State(Cell.Row, Cell.Column) := (Text, Active);
       Claw.Rectangles.Invalidate(Board,
         (Left=>Left(Cell.Column), Right=>Left(Cell.Column+1),
          Top=>Top(Cell.Row), Bottom=>Top(Cell.Row+1)),
         Erase_Background=>False);
    end if;
    if Field.Operations.Game_Over then
       Show_Game_Over;
    end if;
  exception
   when others => null;
  end Display;

  procedure Display_Blank (Cell : in Field.Cell_Location) is
     -- null;
  begin -- Display_Blank
     Display (Cell => Cell, Text => " ", Active => False);
  end Display_Blank;

  procedure Display_Count (Count : in Field.Valid_Count;
                           Stepped : in Boolean;
                           Cell : in Field.Cell_Location) is
    Digit : constant String := "0123456789";
  begin -- Display_Count
    Display (Cell => Cell, Text => (1=>Digit(Count+1)), Active => Stepped);
  end Display_Count;

  procedure Display_Mark (Cell : in Field.Cell_Location) is
  begin
     Display (Cell => Cell, Text => "M", Active => False);
  end Display_Mark;

  procedure Display_Mine (Cell : in Field.Cell_Location) is
  begin
     Display (Cell => Cell, Text => "X", Active => True);
  end Display_Mine;

  procedure Display_To_Go (To_Go : in Integer) is
  begin
    Board.Mines_Left := To_Go;
    Claw.Rectangles.Invalidate(Board, Board.Mines_Left_Area,
                               Erase_Background=>False);
  end Display_To_Go;

  -- screen window creation

  type Menu_Options is (File, New_Game,
                        Options, Auto_Mark, Auto_Step,
                        Help, Rules, About,
                        Quit);

  MENU_BASE : constant Claw.Menu_Identifier_Type :=
      Claw.Menus.Get_Unique_Menu_Identifier (
          Claw.Menu_Item_Index_Type(Menu_Options'Pos(Menu_Options'Last)+1));
          -- This call ensures that these Identifiers are unique in this run.

  procedure Reset_Screen is
     use type Claw.Menu_Identifier_Type;
  begin
     for Row in Field.Valid_Row loop
        for Column in Field.Valid_Column loop
           Board.Button_State (Row, Column) := (" ",False);
        end loop;
     end loop;
     Claw.Menus.Bar.Check_Item(Board.Menu,
            Menu_Options'Pos(Auto_Step) + MENU_BASE,
            Is_Checked=>Field.Operations.Extended_Stepping.Enabled);
     Claw.Menus.Bar.Check_Item(Board.Menu,
            Menu_Options'Pos(Auto_Mark) + MENU_BASE,
            Is_Checked=>Field.Operations.Auto_Marking.Enabled);
     if Board.Ready then
       Claw.Window_Operations.Invalidate(Board);
     end if;
  exception
    when others => null;
  end Reset_Screen;

  procedure Setup(Window : in out Game_Board_Type) is
     use Claw.Menus.Lists;
     use type Claw.Menu_Identifier_Type;
     use type Claw.Int;
     use type Claw.Styles.Window_Style_Type;
     Position : Claw.Rectangle_Type;
     Space_For_Message : constant Claw.Int := Cell_Size;
     One_Char_Size : Claw.Size_Type;
  begin -- Initialization
     Claw.Icons.Create(Fire_Icon, "Fire", (Cell_Size, Cell_Size));
     Claw.Icons.Create(Mark_Icon, "Mark", (Cell_Size, Cell_Size));

     Window.Menu := Create (
       ((1=>Submenu ("&File",
            (1=>Menu_Item ("&New Game", Menu_Options'Pos(New_Game) + MENU_BASE),
             2=>Menu_Item ("E&xit", Menu_Options'Pos(Quit) + MENU_BASE))),
         2=>Submenu ("&Options",
            (1=>Menu_Item ("Auto &Mark", Menu_Options'Pos(Auto_Mark) + MENU_BASE,
                       Kind=>Claw.Menus.Checked, Checked=>False),
             2=>Menu_Item ("Auto &Step", Menu_Options'Pos(Auto_Step) + MENU_BASE,
                       Kind=>Claw.Menus.Checked, Checked=>False))),
         3=>Submenu ("&Help",
            (1=>Menu_Item ("&Rules", Menu_Options'Pos(Rules) + MENU_BASE),
             2=>Menu_Item ("&About", Menu_Options'Pos(About) + MENU_BASE))))));

     for Row in Window.Cell_Top'range loop
       Window.Cell_Top(Row) := Claw.Int((Row-Field.Valid_Row'first)*Cell_Size)
                       + Space_For_Message;
     end loop;
     for Col in Window.Cell_Left'range loop
       Window.Cell_Left(Col) := Claw.Int((Col-Field.Valid_Column'first)*Cell_Size);
     end loop;

     -- find screen position for window to hold game board, menu, etc.
     Position:=(Left=>0, Right=>1+Window.Cell_Left(Window.Cell_Left'last),
                Top=>0, Bottom=>1+Window.Cell_Top(Window.Cell_Top'last));
     Claw.Rectangles.Adjust_For_Style(Position,
            Style=>Claw.Styles.Overlapped_Window,
            Has_Menu=>True);
     Position.Left := Position.Left +5;
     Position.Right := Position.Right +5;
     Position.Bottom := Position.Bottom -Position.Top+10;
     Position.Top := 10;
     Create(Window, "CLAW Mine Detector",
            Style=>Claw.Styles.Overlapped_Window
                   -Claw.Styles.Thick_Frame  -- don't allow resizing
                   -Claw.Styles.Minimize_Box
                   -Claw.Styles.Maximize_Box,
            Position=>Position,
            Menu => Window.Menu,
            Exit_Id => Claw.Menu_Identifier_Type(
                         Menu_Options'Pos(Quit) + MENU_BASE));

     One_Char_Size := Claw.Attributes.Size(Window, 1, Maximum_Size=>True);
     Window.Mines_Left_Area := Claw.Rectangles.To_Rectangle(
         Position=>(X=>Window.Cell_Left(Window.Cell_Left'last-3),
                    Y=>Window.Cell_Top(Window.Cell_Top'first)-One_Char_Size.Height-2),
         Size=> Claw.Attributes.Size(Window,
                  Character_Count=>2, Maximum_Size=>True));
     Window.Game_Over_Area := Claw.Rectangles.To_Rectangle(
         Position=>(X=>Window.Cell_Left(Window.Cell_Left'last)/2
                       -Game_Over_Message'length/2
                       *One_Char_Size.Width,
                    Y=>Window.Cell_Top(Window.Cell_Top'first)-One_Char_Size.Height-1),
         Size=> Claw.Attributes.Size(Window,
                  Game_Over_Message'length, Maximum_Size=>True));

     Field.Operations.Reset;

  end Setup;

  -- handle display of game rules

  Rules_Display : Claw.Edit.Multiline.Multiline_Edit_Type;
  Rule_Box : Rule_Box_Type;

  procedure Display_Rules(Window : in out Game_Board_Type) is
    use type Claw.Styles.Window_Style_Type;
    use type Claw.Edit.Styles.Edit_Style_Type;
    use Claw.Edit.Multiline;
    use Claw.Rectangles;
    Rule_Area : Claw.Rectangle_Type;
  begin
    -- set Rule_Area to center 2/3 of Window's client area
    Rule_Area:=(Top=>Window.Cell_Top(Window.Cell_Top'last/6),
                Bottom=>Window.Cell_Top((5*Window.Cell_Top'last)/6),
                Left=>Window.Cell_Left(Window.Cell_Left'last/6+1),
                Right=>Window.Cell_Left((5*Window.Cell_Left'last)/6+1));
    Create(Rule_Box,"Rules",
        Style=>Claw.Styles.Overlapped_Window+Claw.Styles.Child_Window,
        Position=>Rule_Area,
        Parent=>Window);
    Show(Rule_Box);
    Claw.Edit.Multiline.Create(Rules_Display,Parent=> Rule_Box,
      Style=> Claw.Edit.Styles.Default_Style
              + Claw.Edit.Styles.Read_Only
              + Claw.Styles.Vertical_Scroll_Bar,
      Position=> (0, 0),
      Size=>Claw.Window_Operations.Client_Size(Rule_Box),
      In_Dialog_Units=>False);
    Claw.Edit.Multiline.Set_Text(Rules_Display, Rules_Text);
    Select_Substring(Rules_Display, (1,1)); -- de-select all
    if First_Visible_Line(Rules_Display) > 1 then
      Scroll(Rules_Display,
             Down=> 1-First_Visible_Line(Rules_Display));
    end if;
  exception
    when Claw.Already_Valid_Error => null;  -- just ignore extra request
  end Display_Rules;

  -- if rule box is resized, we must resize the text inside

  procedure When_Resize (Window        : in out Rule_Box_Type;
                         Resize_Params : in     Claw.Codes.Resize_Params_Type;
                         Do_Default    : in out Boolean) is
  begin
    Do_Default := True;
    -- resize contained Rules_Display
    Claw.Window_Operations.Set_Location (Rules_Display,
        Insert_After=>Window, -- place holder, we don't actually change Z order
        Position => (0,0),
        Size => (Width  => Claw.Int(Resize_Params.Client_Width),
                 Height => Claw.Int(Resize_Params.Client_Height)),
        Flags => Claw.Window_Operations.Locations.No_Z_Order);
  exception
    when Claw.Not_Valid_Error => null; -- no child? no op.
  end When_Resize;

  -- handle menu selections

  procedure When_Command (Window : in out Game_Board_Type;
                          From   : in Claw.Command_Source_Type;
                          Command_Id : in Claw.Menu_Identifier_Type;
                          Unknown_Command : in out Boolean) is
    use type Claw.Menu_Identifier_Type;
    use type Claw.Int;
  begin
      Unknown_Command := True; -- Until we know better.
      -- Process a menu selection.
      if Command_Id in
          MENU_BASE .. MENU_BASE + Menu_Options'Pos(Menu_Options'Last) then
          case Menu_Options'Val(Command_ID - MENU_BASE) is
              when Quit =>
                  Unknown_Command := False; -- system will handle it
              when New_Game =>
                  Unknown_Command := False;
                  Field.Operations.Reset;
              when Auto_Mark =>
                  Unknown_Command := False;
                  Field.Operations.Auto_Marking.Enable
                    (Enabled => Not (Field.Operations.Auto_Marking.Enabled));
                  Claw.Menus.Bar.Check_Item(Window.Menu,
                         Menu_Options'Pos(Auto_Mark) + MENU_BASE,
                         Is_Checked=>Field.Operations.Auto_Marking.Enabled);
              when Auto_Step =>
                  Unknown_Command := False;
                  Field.Operations.Extended_Stepping.Enable
                    (Enabled => Not (Field.Operations.Extended_Stepping.Enabled));
                  Claw.Menus.Bar.Check_Item(Window.Menu,
                         Menu_Options'Pos(Auto_Step) + MENU_BASE,
                         Is_Checked=>Field.Operations.Extended_Stepping.Enabled);
              when Rules =>
                  Unknown_Command := False;
                  Display_Rules(Window);
              when About =>
                  Unknown_Command := False;
                  Claw.Message_Box.Message_Box(Window,
                    Text=>About_Text,
                    Caption=>"About Claw Mine Detector",
                    Flags=>Claw.Message_Box.Make_Flags);
              when others   => Unknown_Command:=True;
          end case;
      end if;
  exception
      when others => Unknown_Command:=True;
  end When_Command;

  -- handle drawing of all parts of the Board window

  function Overlap(A,B : Claw.Rectangle_Type) return Boolean is
    X : Claw.Rectangle_Type;
  begin
    Claw.Rectangles.Intersect(Dest=>X, Src1=>A, Src2=>B);
    return not Claw.Rectangles.Is_Empty(X);
  end Overlap;

  procedure When_Draw (Window : in out Game_Board_Type;
                       Easel  : in out Claw.Canvas.Basic_Canvas_Type'Class;
                       Erase_Background : in Boolean;
                       Area   : in Claw.Rectangle_Type) is
    use type Claw.Int;
    use Claw.Colors;
    Pen : Claw.Pens.Pen_Type := Claw.Pens.Create(Claw.Pens.Black);
    Window_Background_Brush : Claw.Brushes.Brush_Type
      := Claw.Brushes.Create(Claw.Attributes.Get_Background_Color(Easel));
    This_Area : Claw.Rectangle_Type;
    Unexplored_Brush : Claw.Brushes.Brush_Type
      := Claw.Brushes.Create(Claw.Brushes.Light_Gray);
    Explored_Brush : Claw.Brushes.Brush_Type
      := Claw.Brushes.Create(Claw.Brushes.Gray);
    Font : Claw.Fonts.Font_Type
      := Claw.Fonts.Create(Font_Family=>Claw.Fonts.Swiss);
  begin
    if not Window.Ready then -- just pass on to parent's When_Draw
      Claw.Frame_Window.When_Draw(Claw.Frame_Window.Frame_Window_Type(Window),
                                  Easel, Erase_Background, Area);
      return;
    end if;

    Claw.Canvas.Choose(Easel, Pen);

    Claw.Canvas.Choose(Easel, Font);

    if Overlap(Area, Window.Mines_Left_Area) then
      Claw.Rectangles.Fill(Easel,Window.Mines_Left_Area,Window_Background_Brush);
      Claw.Draw.Put(Easel,
        Position=> Claw.Rectangles.To_Position(Window.Mines_Left_Area),
        Text => Integer'image(Window.Mines_Left));
    end if;

    if Overlap(Area, Window.Game_Over_Area) then
      Claw.Rectangles.Fill(Easel,Window.Game_Over_Area,Window_Background_Brush);
      if Field.Operations.Game_Over then
        Claw.Draw.Put(Easel,
          Position=> Claw.Rectangles.To_Position(Window.Game_Over_Area),
          Text => Game_Over_Message);
      end if;
    end if;

    Claw.Attributes.Set_Background_Mode (Easel,
                                   Claw.Attributes.Transparent);

    for Row in Field.Valid_Row loop
      for Col in Field.Valid_Column loop
        This_Area
          := (Left=>Window.Cell_Left(Col), Right=>Window.Cell_Left(Col+1),
              Top=>Window.Cell_Top(Row), Bottom=>Window.Cell_Top(Row+1));
        if Overlap(Area,This_Area) then
          if Window.Button_State(Row, Col).Text = "M" then
            --Claw.Canvas.Choose(Easel,Unexplored_Brush);  -- text version
            --Claw.Draw.Rectangle(Easel,This_Area);
            --Claw.Draw.Put(Easel,(X=>Window.Cell_Left(Col)+4, Y=>Window.Cell_Top(Row)+3),"M");
            Claw.Icons.Draw (Mark_Icon, Easel,
                             Start=>(This_Area.Left, This_Area.Top),
                             Size=>(Cell_Size, Cell_Size));
          elsif Window.Button_State(Row, Col).Text = "X" then
            --Claw.Canvas.Choose(Easel,Unexplored_Brush);  -- text version
            --Claw.Draw.Rectangle(Easel,This_Area);
            --Claw.Draw.Put(Easel,(X=>Window.Cell_Left(Col)+4, Y=>Window.Cell_Top(Row)+3),"X");
            Claw.Icons.Draw (Fire_Icon, Easel,
                             Start=>(This_Area.Left, This_Area.Top),
                             Size=>(Cell_Size, Cell_Size));
          else
            if Window.Button_State(Row, Col).Active then
              Claw.Canvas.Choose(Easel,Explored_Brush);
              Claw.Draw.Rectangle(Easel,This_Area);
            else
              Claw.Canvas.Choose(Easel,Unexplored_Brush);
              Claw.Draw.Rectangle(Easel,This_Area);
            end if;
            Claw.Draw.Put(Easel,
                 Position=> (X=>Window.Cell_Left(Col)+6, Y=>Window.Cell_Top(Row)+1),
                 Text => Window.Button_State(Row, Col).Text);
          end if;
        end if;
      end loop;
    end loop;

  end When_Draw;

  -- handle mouse clicks

  procedure When_Left_Button_Down
                           (Window       : in out Game_Board_Type;
                            Mouse_Params : in     Claw.Mouse.Mouse_Params_Type;
                            Do_Default   : in out Boolean) is
    use type Claw.Int;
    Row, Col : Integer;
    Cell_Top : Position_Lists renames Window.Cell_Top;
    Cell_Left : Position_Lists renames Window.Cell_Left;
  begin
    if not Window.Ready then -- ignore premature mousing
      return;
    end if;
    if Field.Operations.Game_Over then
       Show_Game_Over;
    elsif Claw.Int(Mouse_Params.Y)
          in Cell_Top(Cell_Top'first) .. Cell_Top(Cell_Top'last)-1
    and Claw.Int(Mouse_Params.X)
        in Cell_Left(Cell_Left'first) .. Cell_Left(Cell_Left'last)-1 then
      Row := (Integer(Mouse_Params.Y)-Integer(Cell_Top(Cell_Top'first)))
             /Cell_Size+1;
      Col := (Integer(Mouse_Params.X)-Integer(Cell_Left(Cell_Left'first)))
             /Cell_Size+1;
      Field.Operations.Step (Cell => (Row, Col));
    end if;
  end When_Left_Button_Down;

  procedure When_Right_Button_Down
                           (Window       : in out Game_Board_Type;
                            Mouse_Params : in     Claw.Mouse.Mouse_Params_Type;
                            Do_Default   : in out Boolean) is
    use type Claw.Int;
    Row, Col : Integer;
    Cell_Top : Position_Lists renames Window.Cell_Top;
    Cell_Left : Position_Lists renames Window.Cell_Left;
  begin
    if not Window.Ready then -- ignore premature mousing
      return;
    end if;
    if Field.Operations.Game_Over then
       Show_Game_Over;
    elsif Claw.Int(Mouse_Params.Y)
          in Cell_Top(Cell_Top'first) .. Cell_Top(Cell_Top'last)-1
    and Claw.Int(Mouse_Params.X)
        in Cell_Left(Cell_Left'first) .. Cell_Left(Cell_Left'last)-1 then
      Row := (Integer(Mouse_Params.Y)-Integer(Cell_Top(Cell_Top'first)))
             /Cell_Size+1;
      Col := (Integer(Mouse_Params.X)-Integer(Cell_Left(Cell_Left'first)))
             /Cell_Size+1;
      Field.Operations.Mark (Cell => (Row, Col));
    end if;
  end When_Right_Button_Down;

  procedure Run is
  begin
    Setup(Board);
    Board.Ready := True;  -- everything is ready, display the board!
    Show(Board, How=>Claw.Codes.Show_Startup);
    -- Now let the main program thread just sit till the user closes
    -- the Board
    Wait_for_Close (Window => Board);
  end Run;

end User_If;
--
-- This is free software; you can redistribute it and/or modify it under
-- terms of the GNU General Public License as published by the Free Software
-- Foundation; either version 2, or (at your option) any later version.
-- This software is distributed in the hope that it will be useful, but WITH
-- OUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
-- for more details. Free Software Foundation, 59 Temple Place - Suite
-- 330, Boston, MA 02111-1307, USA.
